import { __metadata, __decorate } from 'tslib';
import { NgxRxdbUtils, isValidRxReplicationState, mapFindResultToJsonArray } from '@ngx-odm/rxdb/utils';
import { RXJS_SHARE_REPLAY_DEFAULTS, removeRxDatabase } from 'rxdb';
import { Observable, switchMap, defer, lastValueFrom, of, ReplaySubject, takeWhile, fromEvent, shareReplay, isObservable, map, distinctUntilChanged } from 'rxjs';

const { debug: debug$1, isEmptyObject, isFunction } = NgxRxdbUtils;
function isZone(obj) {
    return !isEmptyObject(obj) && isFunction(obj.run);
}
function runInZone(zone) {
    if (!isZone(zone))
        return source => source;
    return source => {
        return new Observable(subscriber => {
            return source.subscribe((value) => zone.run(() => subscriber.next(value)), (e) => zone.run(() => subscriber.error(e)), () => zone.run(() => subscriber.complete()));
        });
    };
}
function ensureCollection$() {
    return function (target, propertyKey, descriptor) {
        const originalMethod = descriptor.value;
        descriptor.value = function (...args) {
            return defer(() => {
                return this.initialized$.pipe(switchMap(() => originalMethod.apply(this, args)), debug$1(`collection.${propertyKey}`));
            });
        };
        return descriptor;
    };
}
function ensureCollection() {
    return function (target, propertyKey, descriptor) {
        const originalMethod = descriptor.value;
        descriptor.value = async function (...args) {
            await lastValueFrom(this.initialized$).catch(() => {
                throw new Error(`Collection "${this.config.name}" was not initialized. Please check RxDB errors.`);
            });
            return originalMethod.apply(this, args);
        };
        return descriptor;
    };
}

var _a, _b, _c, _d, _e, _f;
const { getMaybeId, logger, debug, noop } = NgxRxdbUtils;
class RxDBCollectionService {
    get initialized$() {
        return this._init$.asObservable();
    }
    get collection() {
        return this._collection;
    }
    get db() {
        return this.dbService.db;
    }
    get dbOptions() {
        return this.dbService.dbOptions;
    }
    get replicationState() {
        return this._replicationState;
    }
    get queryParams$() {
        if (!this.config.options?.useQueryParams) {
            return of({});
        }
        return this.initialized$.pipe(switchMap(() => this.collection.queryParams.$));
    }
    constructor(config, dbService, ngZone = {}, currentUrl$ = of(''), updateQueryParamsFn = noop) {
        this.config = config;
        this.dbService = dbService;
        this.ngZone = ngZone;
        this.currentUrl$ = currentUrl$;
        this.updateQueryParamsFn = updateQueryParamsFn;
        this._replicationState = null;
        this._init$ = new ReplaySubject();
        this.init(config);
    }
    destroy() {
        this.collection?.destroy();
    }
    setQueryParams(query) {
        this.collection.queryParams?.set(query);
    }
    patchQueryParams(query) {
        this.collection.queryParams?.patch(query);
    }
    async sync() {
        if (isValidRxReplicationState(this.replicationState)) {
            this.replicationState.reSync();
            return;
        }
        if (typeof this.config.options?.replicationStateFactory !== 'function') {
            return;
        }
        try {
            this._replicationState = this.config.options.replicationStateFactory(this.collection);
        }
        catch (error) {
            logger.log('replicationState has error, ignore replication');
            logger.log(error.message);
        }
        if (isValidRxReplicationState(this.replicationState)) {
            if (!this.replicationState.autoStart) {
                this.replicationState.reSync();
            }
            fromEvent(window, 'online')
                .pipe(debug('online'), takeWhile(() => !this.replicationState.isStopped()))
                .subscribe(() => {
                this._replicationState.reSync();
            });
            return this.replicationState.startPromise;
        }
    }
    async info() {
        return this.collection.getMetadata();
    }
    async import(docs) {
        const schemaHash = await this.collection.schema.hash;
        const dump = {
            name: this.collection.name,
            schemaHash,
            docs,
        };
        this.collection.importJSON(dump);
    }
    async export() {
        return this.collection.exportJSON();
    }
    docs(query, withRevAndAttachments = false) {
        return (isObservable(query) ? query : of(query)).pipe(switchMap(q => this.collection.find(q).$), mapFindResultToJsonArray(withRevAndAttachments), runInZone(this.ngZone), shareReplay(RXJS_SHARE_REPLAY_DEFAULTS));
    }
    docsByIds(ids, withRevAndAttachments = false) {
        return this.collection.findByIds(ids).$.pipe(mapFindResultToJsonArray(withRevAndAttachments), runInZone(this.ngZone), shareReplay(RXJS_SHARE_REPLAY_DEFAULTS));
    }
    count(query) {
        return this.collection.count(query).$.pipe(runInZone(this.ngZone), shareReplay(RXJS_SHARE_REPLAY_DEFAULTS));
    }
    get(id, withRevAndAttachments = false) {
        return this.collection.findOne(id).$.pipe(map(doc => (doc ? doc.toMutableJSON(withRevAndAttachments) : null)), runInZone(this.ngZone), shareReplay(RXJS_SHARE_REPLAY_DEFAULTS));
    }
    async insert(data) {
        return this.collection.insert(data);
    }
    async insertBulk(data) {
        return this.collection.bulkInsert(data);
    }
    async upsert(data) {
        return this.collection.upsert(data);
    }
    async upsertBulk(data) {
        return this.collection.bulkUpsert(data);
    }
    async set(id, data) {
        return this.collection.findOne(id).update({ $set: data });
    }
    async updateBulk(query, data) {
        return this.collection.find(query).update({ $set: data });
    }
    async remove(entityOrId) {
        const id = getMaybeId(entityOrId);
        return this.collection.findOne(id).remove();
    }
    async removeBulk(params) {
        if (Array.isArray(params)) {
            return this.collection.bulkRemove(params);
        }
        const result = await this.collection
            .find(params)
            .remove()
            .catch(() => new Map());
        return {
            success: [...result.values()],
            error: [],
        };
    }
    async clear() {
        return this.collection.remove();
    }
    async getAttachments(docId) {
        const doc = await this.collection.findOne(docId).exec();
        if (!doc) {
            return null;
        }
        const attachmentsData = doc.allAttachments().map(a => a.getData());
        return Promise.all(attachmentsData);
    }
    async getAttachmentById(docId, attachmentId) {
        const doc = await this.collection.findOne(docId).exec();
        if (!doc) {
            return null;
        }
        const attachment = doc.getAttachment(attachmentId);
        if (!attachment) {
            return null;
        }
        return attachment.getData();
    }
    async putAttachment(docId, attachment) {
        const doc = await this.collection.findOne(docId).exec();
        if (!doc) {
            return;
        }
        await doc.putAttachment(attachment);
    }
    async removeAttachment(docId, attachmentId) {
        const doc = await this.collection.findOne(docId).exec();
        if (!doc) {
            return;
        }
        const attachment = doc.getAttachment(attachmentId);
        if (!attachment) {
            return;
        }
        await attachment.remove();
    }
    async addHook(hook, handler, parralel = false) {
        this.collection[hook](handler, parralel);
    }
    async getLocal(id, key) {
        const doc = await this.collection.getLocal(id);
        if (!doc) {
            return null;
        }
        return key ? doc?.get(key) : doc?.toJSON().data;
    }
    getLocal$(id, key) {
        return this.collection.getLocal$(id).pipe(map(doc => {
            if (!doc) {
                return null;
            }
            return key ? doc.get(key) : doc.toJSON().data;
        }), distinctUntilChanged(), runInZone(this.ngZone));
    }
    async insertLocal(id, data) {
        await this.collection.insertLocal(id, data);
    }
    async upsertLocal(id, data) {
        await this.collection.upsertLocal(id, data);
    }
    async setLocal(id, prop, value) {
        const loc = await this.collection.getLocal(id);
        await this.collection.upsertLocal(id, {
            ...(loc?.toJSON?.()?.data || {}),
            [prop]: value,
        });
    }
    async removeLocal(id) {
        const doc = await this.collection.getLocal(id);
        await doc?.remove();
    }
    async init(config) {
        const { name, options } = config;
        try {
            await this.dbService.initCollections({
                [name]: config,
            });
            this._collection = this.db.collections[name];
            if (options?.useQueryParams) {
                this.collection.queryParamsInit(this.currentUrl$, this.updateQueryParamsFn);
            }
            this._init$.next(true);
            this._init$.complete();
        }
        catch (err) {
            if (err.message.includes('DM5') ||
                (err.message.includes('DB6') &&
                    err.parameters.previousSchema?.version ===
                        err.parameters.schema?.version)) {
                logger.log('Reload the page to fix the issue. The database is in a state where it can not be used.');
                await removeRxDatabase(this.db.name, this.db.storage);
                window?.location?.reload?.();
            }
            else {
                this._init$.complete();
                throw err;
            }
        }
    }
}
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "sync", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "info", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Array]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "import", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "export", null);
__decorate([
    ensureCollection$(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object, Object]),
    __metadata("design:returntype", Observable)
], RxDBCollectionService.prototype, "docs", null);
__decorate([
    ensureCollection$(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Array, Object]),
    __metadata("design:returntype", Observable)
], RxDBCollectionService.prototype, "docsByIds", null);
__decorate([
    ensureCollection$(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", Observable)
], RxDBCollectionService.prototype, "count", null);
__decorate([
    ensureCollection$(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object]),
    __metadata("design:returntype", Observable)
], RxDBCollectionService.prototype, "get", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "insert", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Array]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "insertBulk", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "upsert", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Array]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "upsertBulk", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "set", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object, Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "updateBulk", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "remove", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "removeBulk", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "clear", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "getAttachments", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, String]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "getAttachmentById", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "putAttachment", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, String]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "removeAttachment", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [typeof (_a = typeof Hook !== "undefined" && Hook) === "function" ? _a : Object, Object, Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "addHook", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, typeof (_b = typeof K !== "undefined" && K) === "function" ? _b : Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "getLocal", null);
__decorate([
    ensureCollection$(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, typeof (_c = typeof K !== "undefined" && K) === "function" ? _c : Object]),
    __metadata("design:returntype", Observable)
], RxDBCollectionService.prototype, "getLocal$", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, typeof (_d = typeof L !== "undefined" && L) === "function" ? _d : Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "insertLocal", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, typeof (_e = typeof L !== "undefined" && L) === "function" ? _e : Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "upsertLocal", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, typeof (_f = typeof K !== "undefined" && K) === "function" ? _f : Object, Object]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "setLocal", null);
__decorate([
    ensureCollection(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", Promise)
], RxDBCollectionService.prototype, "removeLocal", null);

export { RxDBCollectionService };
//# sourceMappingURL=ngx-odm-rxdb-collection.mjs.map
