import { NgxRxdbUtils, getDefaultFetch } from '@ngx-odm/rxdb/utils';
import { compare } from 'compare-versions';
import { isRxDatabaseFirstTimeInstantiated, getAllCollectionDocuments } from 'rxdb';

const { logger, getDefaultPreparedQuery } = NgxRxdbUtils;
const RXDB_STORAGE_TOKEN_ID = 'storage-token|storageToken';
const fetchSchema = async (schemaUrl) => {
    let schema;
    try {
        const fetch = getDefaultFetch();
        const result = await fetch(schemaUrl);
        if (result.ok) {
            schema = await result.json();
        }
        else {
            throw new Error(`Failed to fetch schema from "${schemaUrl}", status: ${result.status}`);
        }
    }
    catch (error) {
        throw new Error(`Failed to fetch schema from "${schemaUrl}"`);
    }
    return schema;
};
const prepareCollections = async (colConfigs) => {
    try {
        const colCreators = {};
        for (const name in colConfigs) {
            const config = colConfigs[name];
            if (!config.schema && !!config.options?.schemaUrl) {
                config.schema = (await fetchSchema(config.options.schemaUrl));
            }
            colCreators[config.name] = config;
        }
        return colCreators;
    }
    catch (error) {
        throw new Error(error.message);
    }
};
const prepareDbDump = async (dumpPath, collections) => {
    let dumpObj;
    const fetch = getDefaultFetch();
    const result = await fetch(dumpPath);
    if (result.ok) {
        dumpObj = await result.json();
    }
    else {
        throw new Error(`Failed to fetch db dump from "${dumpPath}", status: ${result.status}`);
    }
    if (!collections || !Object.keys(collections).length) {
        throw new Error('collections must be initialized before importing');
    }
    for (const dc of dumpObj.collections) {
        const col = collections[dc.name];
        if (col) {
            dc.schemaHash = col.schema['_hash'];
        }
        else {
            throw new Error('no such collection as provided in dump');
        }
    }
    return dumpObj;
};
const migrateStorageVersion = async (storage, storageInstance, rxdbVersion) => {
    let storageData = {};
    if (storage.name !== 'dexie') {
        return;
    }
    const dexieTable = (await storageInstance.internals)?.dexieTable;
    storageData = (await dexieTable.get(RXDB_STORAGE_TOKEN_ID))?.data;
    if (!storageData?.rxdbVersion || compare(rxdbVersion, storageData.rxdbVersion, '>')) {
        try {
            await dexieTable.update(RXDB_STORAGE_TOKEN_ID, {
                data: {
                    ...storageData,
                    rxdbVersion,
                },
            });
            logger.log('prepare-plugin: migrated internal storage to', rxdbVersion);
        }
        catch (error) {
            throw new Error('prepare-plugin: unable to migrate internal storage');
        }
    }
    return;
};
const afterCreateRxDatabase = async ({ database: db, creator, }) => {
    const isFirstTimeInstantiated = await isRxDatabaseFirstTimeInstantiated(db);
    logger.log('prepare-plugin: hook:createRxDatabase:after');
    const { storage, internalStore: storageInstance, rxdbVersion } = db;
    await migrateStorageVersion(storage, storageInstance, rxdbVersion);
    if (!creator.options?.dumpPath || !isFirstTimeInstantiated) {
        return;
    }
    try {
        const dump = await prepareDbDump(creator.options.dumpPath, db.collections);
        await db.importJSON(dump);
        logger.log(`prepare-plugin: imported dump for db "${db.name}"`);
    }
    catch (error) {
        logger.log('prepare-plugin: imported dump error', error);
    }
};
const beforePreCreateRxSchema = async (maybeSchema) => {
    if (typeof maybeSchema === 'string') {
        const realSchema = await fetchSchema(maybeSchema);
        if (!realSchema) {
            throw new Error(`Failed to fetch schema from "${maybeSchema}"`);
        }
        maybeSchema = realSchema;
    }
};
const afterCreateRxCollection = async ({ collection: col, creator, }) => {
    logger.log('prepare-plugin: hook:createRxCollection:before');
    const meta = await col.getMetadata();
    logger.log('prepare-plugin: hook:createRxCollection:before', meta);
    const initialDocs = creator.options?.initialDocs || [];
    const initialCount = await countDocs();
    if (!initialDocs.length) {
        return;
    }
    if (initialCount || !meta.isFirstTimeInstantiated) {
        if (!creator.options?.recreate || creator.options?.replication) {
            return;
        }
        else {
            logger.log(`prepare-plugin: collection "${col.name}" already has ${initialCount} docs (including _deleted), but recreate option is set`);
        }
    }
    const schemaHash = await col.schema.hash;
    const dump = {
        name: col.name,
        schemaHash,
        docs: initialDocs,
    };
    try {
        const { success, error } = (await col.importJSON(dump));
        const count = await countDocs();
        logger.log(`prepare-plugin: imported ${success.length} docs for collection "${col.name}", errors count ${error.length}, current docs count ${count}`);
        logger.table(success);
    }
    catch (error) {
        logger.log('prepare-plugin: imported dump error', error);
    }
    return;
    async function removeAllDocs() {
        await col.find().update({ $set: { _deleted: true } });
        await col.cleanup();
    }
    async function countDocs() {
        const { count } = await col.storageInstance
            .count(getDefaultPreparedQuery())
            .catch(() => ({ count: 0 }));
        return count;
    }
};
const RxDBPreparePlugin = {
    name: 'prepare-plugin',
    rxdb: true,
    prototypes: {
        RxDatabase: (proto) => {
            Object.assign(proto, {
                fetchSchema,
            });
        },
        RxCollection: (proto) => {
            const getMetadata = async function () {
                const isFirstTimeInstantiated = await isRxDatabaseFirstTimeInstantiated(this.database);
                const allCollectionMetaDocs = await getAllCollectionDocuments(this.database.internalStore);
                const { id, data, _meta, _rev } = allCollectionMetaDocs.filter(metaDoc => metaDoc.data.name === this.name).at(0) ||
                    {};
                return {
                    id: id || this.name,
                    databaseName: this.database.name,
                    collectionName: data?.name || this.name,
                    storageName: this.storageInstance.originalStorageInstance['storage'].name,
                    last_modified: _meta?.lwt ? Math.floor(_meta?.lwt) : Date.now(),
                    rev: _rev ? Number(_rev?.at(0)) : 1,
                    isFirstTimeInstantiated,
                };
            };
            Object.assign(proto, {
                getMetadata,
            });
        },
    },
    hooks: {
        createRxDatabase: {
            after: afterCreateRxDatabase,
        },
        preCreateRxSchema: {
            before: beforePreCreateRxSchema,
        },
        createRxCollection: {
            after: afterCreateRxCollection,
        },
    },
};

export { RxDBPreparePlugin, beforePreCreateRxSchema, fetchSchema, prepareCollections };
//# sourceMappingURL=ngx-odm-rxdb-prepare.mjs.map
