import { NgxRxdbUtils } from '@ngx-odm/rxdb/utils';
import { BehaviorSubject, distinctUntilChanged, tap, map, catchError, takeWhile, EMPTY } from 'rxjs';
import { includeKeys } from 'filter-obj';
import queryString from 'query-string';

const { keys, isEmpty, isNullOrUndefined } = NgxRxdbUtils;
const normalizeMangoQuery = ({ selector, sort, limit, skip }, schema) => {
    const schemaTopLevelFields = keys(schema.jsonSchema.properties);
    if (!isEmpty(selector)) {
        selector = includeKeys(selector, key => {
            return (!key.startsWith('$') &&
                !key.includes('.') &&
                schemaTopLevelFields.includes(key));
        });
    }
    if (!isEmpty(sort)) {
        sort = sort?.filter(sortPart => schemaTopLevelFields.includes(keys(sortPart)[0]));
    }
    if (isNullOrUndefined(limit) || isNaN(limit)) {
        limit = undefined;
    }
    if (isNullOrUndefined(skip) || isNaN(skip)) {
        skip = undefined;
    }
    return { selector, sort, limit, skip };
};
const parseUrlToMangoQuery = (url, schema) => {
    url = queryString.extract(url);
    const urlPart = queryString.pick(url, ['selector', 'sort', 'limit', 'skip']);
    const parsed = queryString.parse(urlPart, {
        parseNumbers: true,
        parseBooleans: true,
    });
    const { selector: _selector, sort: _sort, limit, skip } = parsed;
    const selector = _selector ? JSON.parse(_selector) : undefined;
    const sort = _sort ? JSON.parse(_sort) : undefined;
    const queryObj = normalizeMangoQuery({
        selector,
        sort,
        limit,
        skip,
    }, schema);
    return queryObj;
};
const stringifyParam = (param) => {
    if (isNullOrUndefined(param))
        return '';
    return JSON.stringify(param);
};

const { logger, compactObject } = NgxRxdbUtils;
const RxDBPUseQueryParamsPlugin = {
    name: 'query-params-plugin',
    rxdb: true,
    prototypes: {
        RxCollection: (proto) => {
            const _queryParams$ = new BehaviorSubject({});
            let useQueryParams = false;
            let schema;
            let initDone = false;
            let currentUrl = '';
            let updateQueryParamsInLocationFn = () => Promise.resolve(false);
            function init(currentUrl$ = EMPTY, _updateQueryParamsInLocationFn) {
                useQueryParams = !!this.options?.useQueryParams;
                schema = this.schema;
                if (!useQueryParams || initDone)
                    return;
                updateQueryParamsInLocationFn = _updateQueryParamsInLocationFn;
                currentUrl$
                    .pipe(distinctUntilChanged(), tap(url => {
                    currentUrl = url;
                }), map(url => parseUrlToMangoQuery(url, schema)), catchError(err => {
                    logger.log('Error in parsing url to mango query', err);
                    return _queryParams$;
                }), tap(queryParams => {
                    logger.log('queryParams', queryParams);
                }), takeWhile(() => !this.destroyed))
                    .subscribe(_queryParams$);
                initDone = true;
            }
            function set(query) {
                if (!useQueryParams)
                    return;
                const { selector, sort, limit, skip } = query;
                const queryParams = { limit, skip };
                if (selector) {
                    queryParams.selector = stringifyParam(query.selector);
                }
                if (sort) {
                    queryParams.sort = stringifyParam(query.sort);
                }
                updateQueryParamsInLocationFn(queryParams);
            }
            function patch(query) {
                if (!useQueryParams)
                    return;
                const parsed = parseUrlToMangoQuery(currentUrl, schema);
                const queryParams = {
                    selector: stringifyParam(query.selector || parsed.selector),
                    sort: stringifyParam(query.sort || parsed.sort),
                    limit: query.limit || parsed.limit,
                    skip: query.skip || parsed.skip,
                };
                updateQueryParamsInLocationFn(compactObject(queryParams));
            }
            Object.assign(proto, {
                queryParams: {
                    $: _queryParams$.asObservable(),
                    get: () => _queryParams$.getValue(),
                    set,
                    patch,
                },
            }, {
                queryParamsInit: init,
            });
        },
    },
    hooks: {
        createRxCollection: {},
    },
};

export { RxDBPUseQueryParamsPlugin };
//# sourceMappingURL=ngx-odm-rxdb-query-params.mjs.map
