import { NgxRxdbUtils, getDefaultFetch } from '@ngx-odm/rxdb/utils';
import { addRxPlugin } from 'rxdb';
import { RxDBLeaderElectionPlugin } from 'rxdb/plugins/leader-election';
import { RxReplicationState, startReplicationOnLeaderShip } from 'rxdb/plugins/replication';
import { Subject, EMPTY, switchMap, map, of, first, lastValueFrom, interval, takeWhile, from, filter, merge } from 'rxjs';

var KintoReplicationStrategy;
(function (KintoReplicationStrategy) {
    KintoReplicationStrategy["CLIENT_WINS"] = "CLIENT_WINS";
    KintoReplicationStrategy["SERVER_WINS"] = "SERVER_WINS";
    KintoReplicationStrategy["PULL_ONLY"] = "PULL_ONLY";
    KintoReplicationStrategy["MANUAL"] = "MANUAL";
})(KintoReplicationStrategy || (KintoReplicationStrategy = {}));

const DEFAULT_KINTO_SYNC_OPTIONS = {
    remote: '/',
    bucket: 'default',
    collection: 'default',
    headers: {},
    exclude: undefined,
    expectedTimestamp: undefined,
    timeout: undefined,
    heartbeat: 60000,
    strategy: KintoReplicationStrategy.CLIENT_WINS,
};
const DEFAULT_REPLICATION_OPTIONS = {
    kintoSyncOptions: DEFAULT_KINTO_SYNC_OPTIONS,
    live: true,
    retryTime: 1000 * 5,
    waitForLeadership: true,
    autoStart: true,
};
function assignLastModified(doc, last_modified, force = false) {
    if (!doc.last_modified || force) {
        doc.last_modified = last_modified;
    }
    return doc;
}
function addEndpointOptions(path, options = {}) {
    const query = { ...options.query };
    if (options.fields) {
        query._fields = options.fields;
    }
    const queryString = qsify(query);
    if (queryString) {
        return path + '?' + queryString;
    }
    return path;
}
function cleanUndefinedProperties(obj) {
    const result = {};
    for (const key in obj) {
        if (typeof obj[key] !== 'undefined') {
            result[key] = obj[key];
        }
    }
    return result;
}
function qsify(obj) {
    const encode = (v) => encodeURIComponent(typeof v === 'boolean' ? String(v) : v);
    const stripped = cleanUndefinedProperties(obj);
    return Object.keys(stripped)
        .map(k => {
        const ks = encode('_' + k) + '=';
        if (Array.isArray(stripped[k])) {
            return ks + stripped[k].map((v) => encode(v)).join(',');
        }
        return ks + encode(stripped[k]);
    })
        .join('&');
}
function aggregate(responses = [], requests = []) {
    if (responses.length !== requests.length) {
        throw new Error('Responses length should match requests one.');
    }
    let last_modified;
    const results = {
        errors: [],
        published: [],
        conflicts: [],
        skipped: [],
        last_modified,
    };
    return responses.reduce((acc, response, index) => {
        const { status, headers } = response;
        const etag = headers?.['ETag'];
        acc.last_modified = etag ? +etag.replace(/"/g, '') : undefined;
        const request = requests[index];
        if (status >= 200 && status < 400) {
            acc.published.push(response.body);
        }
        else if (status === 404) {
            const regex = /(buckets|groups|collections|records)\/([^/]+)$/;
            const extracts = request.path.match(regex);
            const id = extracts && extracts.length === 3 ? extracts[2] : undefined;
            acc.skipped.push({
                id,
                path: request.path,
                error: response.body,
            });
        }
        else if (status === 412) {
            acc.conflicts.push({
                type: 'outgoing',
                local: request.body,
                remote: (response.body.details && response.body.details.existing) || null,
            });
        }
        else {
            acc.errors.push({
                path: request.path,
                sent: request,
                error: response.body,
            });
        }
        return acc;
    }, results);
}
function mergeAggregatedResponse(obj, next) {
    for (const key in next) {
        obj[key] = next[key];
    }
    return obj;
}
function kintoCollectionFactory({ remote, bucket, collection: collectionName, headers, exclude, expectedTimestamp, } = DEFAULT_KINTO_SYNC_OPTIONS, fetch = getDefaultFetch()) {
    const collectionUrl = `buckets/${bucket}/collections/${collectionName}`;
    let filters;
    if (exclude) {
        const exclude_id = exclude
            .slice(0, 50)
            .map(r => r.id)
            .join(',');
        filters = { exclude_id };
    }
    if (expectedTimestamp) {
        filters = {
            ...filters,
            _expected: expectedTimestamp,
        };
    }
    return {
        async info() {
            const response = await fetch(`${remote}/${collectionUrl}`, { headers });
            return response.json();
        },
        async listRecords(options) {
            const { ...query } = options;
            if (filters) {
                query.filters = filters;
            }
            if (!query.sort) {
                query.sort = '-last_modified';
            }
            const url = addEndpointOptions(`${remote}/${collectionUrl}/records`, { query });
            const response = await fetch(url, { headers });
            const nextPage = response.headers.get('Next-Page');
            const etag = response.headers.get('ETag');
            const last_modified = etag ? etag.replace(/"/g, '') : etag;
            const { data } = await response.json();
            NgxRxdbUtils.logger.log('replication:kinto:since', response.status, last_modified);
            return {
                data,
                hasNextPage: !!nextPage,
                last_modified,
            };
        },
        async batch(changes) {
            if (!changes.length) {
                return {
                    published: [],
                    errors: [],
                    conflicts: [],
                    skipped: [],
                    last_modified: undefined,
                };
            }
            const toDelete = [];
            const toUpdate = [];
            const toCreate = [];
            changes.forEach(doc => {
                if (doc.deleted) {
                    toDelete.push(doc);
                }
                else if (doc.last_modified) {
                    toUpdate.push(doc);
                }
                else {
                    toCreate.push(doc);
                }
            });
            const body = {
                defaults: { headers },
                requests: [
                    ...toCreate.map(({ ...data }) => ({
                        method: 'PUT',
                        path: `/${collectionUrl}/records/${data.id}`,
                        headers: { 'If-None-Match': '*' },
                        body: { data },
                    })),
                    ...toUpdate.map(({ last_modified, ...data }) => ({
                        method: 'PATCH',
                        path: `/${collectionUrl}/records/${data.id}`,
                        headers: { 'If-Match': `"${last_modified}"` },
                        body: { data },
                    })),
                    ...toDelete.map(({ id }) => ({
                        method: 'DELETE',
                        path: `/${collectionUrl}/records/${id}`,
                    })),
                ],
            };
            const bodyString = JSON.stringify(body);
            const response = await fetch(`${remote}/batch`, {
                headers,
                method: 'POST',
                body: bodyString,
            });
            NgxRxdbUtils.logger.log('replication:kinto:batch', response.status);
            const { responses, requests } = await response.json();
            const result = aggregate(responses, requests || body.requests);
            return result;
        },
    };
}

function replicateKintoDB(options) {
    const { replicationIdentifier, kintoSyncOptions, fetch, pull, push, collection, retryTime: retry, live, autoStart, waitForLeadership, } = Object.assign(DEFAULT_REPLICATION_OPTIONS, options);
    addRxPlugin(RxDBLeaderElectionPlugin);
    const kintoCollection = kintoCollectionFactory(kintoSyncOptions, fetch);
    const pullStream$ = new Subject();
    let _pullImplementation;
    let _pushImplementation;
    let _pullHandlerResult$ = EMPTY;
    let since;
    if (pull) {
        _pullImplementation = {
            async handler(lastPulledCheckpoint, batchSize) {
                since = lastPulledCheckpoint?.last_modified || undefined;
                const limit = batchSize || undefined;
                _pullHandlerResult$ = of(null).pipe(switchMap(() => kintoCollection.listRecords({ since, limit })), map(({ data: documents, last_modified, hasNextPage }) => {
                    since = last_modified;
                    return {
                        documents,
                        checkpoint: { last_modified },
                    };
                }));
                return lastValueFrom(_pullHandlerResult$.pipe(first(), NgxRxdbUtils.retryWithBackoff(1, retry)));
            },
            batchSize: pull?.batchSize,
            modifier: pull?.modifier,
            stream$: pullStream$.asObservable(),
        };
    }
    if (push) {
        _pushImplementation = {
            async handler(changes) {
                const { last_modified: lwt } = await collection.getMetadata();
                const { data: { last_modified }, } = await kintoCollection.info();
                const outgoing = changes.map(({ newDocumentState: doc }) => doc);
                const results = {
                    errors: [],
                    published: [],
                    conflicts: [],
                    skipped: [],
                    last_modified,
                };
                const r = await kintoCollection.batch(outgoing);
                mergeAggregatedResponse(results, r);
                if (results.errors.length) {
                    NgxRxdbUtils.logger.log(results.errors);
                }
                if (results.skipped.length) {
                    const missing = results.skipped
                        .filter(({ error }) => error.code == 404)
                        .map(({ path }) => {
                        const doc = outgoing.find((d) => String(path).endsWith(d['id']));
                        if (doc) {
                            delete doc['last_modified'];
                        }
                        return doc;
                    });
                    if (missing.length) {
                        const r = await kintoCollection.batch(missing);
                        mergeAggregatedResponse(results, r);
                    }
                }
                if (results.conflicts.length) {
                    const conflicted = results.conflicts
                        .filter(({ remote }) => !remote)
                        .map(c => {
                        delete c.local.data.last_modified;
                        return c.local.data;
                    });
                    if (conflicted.length) {
                        const r = await kintoCollection.batch(conflicted);
                        mergeAggregatedResponse(results, r);
                    }
                }
                if (results.published.length) {
                    return (results.published
                        .filter(({ data }) => !data.deleted)
                        .map(({ data }) => data));
                }
                return results.conflicts.map(c => c.remote);
            },
            batchSize: push?.batchSize,
            modifier: push?.modifier,
        };
    }
    const replication = new RxReplicationState(replicationIdentifier, collection, 'deleted', _pullImplementation, _pushImplementation, live, retry, autoStart);
    if (live && pull) {
        merge(from(replication.awaitInitialReplication()), replication.remoteEvents$.pipe(filter(e => e === 'RESYNC')))
            .pipe(switchMap(() => interval(pull.heartbeat || 60000)), switchMap(() => _pullHandlerResult$), NgxRxdbUtils.debug('replication:kinto:longpoll'), NgxRxdbUtils.retryWithBackoff(3, retry), takeWhile(() => !replication.isStopped()))
            .subscribe(pullStream$);
    }
    startReplicationOnLeaderShip(true, replication);
    return replication;
}

export { KintoReplicationStrategy, replicateKintoDB };
//# sourceMappingURL=ngx-odm-rxdb-replication-kinto.mjs.map
