import { Injector, inject, computed } from '@angular/core';
import { getCallStateKeys, setLoading, setLoaded, setError } from '@angular-architects/ngrx-toolkit';
import { tapResponse } from '@ngrx/operators';
import { withState, withComputed, patchState, withMethods, withHooks, signalStoreFeature } from '@ngrx/signals';
import { setAllEntities } from '@ngrx/signals/entities';
import { RXDB_COLLECTION } from '@ngx-odm/rxdb';
import { RxDBCollectionService } from '@ngx-odm/rxdb/collection';
import { DEFAULT_LOCAL_DOCUMENT_ID } from '@ngx-odm/rxdb/config';
import { NgxRxdbUtils } from '@ngx-odm/rxdb/utils';
import { computedAsync } from 'ngxtension/computed-async';
import { tap, switchMap } from 'rxjs';

const { isEmpty, logger, tapOnce } = NgxRxdbUtils;
function capitalize(str) {
    return str ? str[0].toUpperCase() + str.substring(1) : str;
}
const queryKey = 'query';
const countKey = 'count';
const countAllKey = 'countAll';
function getCollectionServiceKeys(options) {
    const filterKey = options.collection ? `${options.collection}Filter` : 'filter';
    const selectedIdsKey = options.collection
        ? `selected${capitalize(options.collection)}Ids`
        : 'selectedIds';
    const selectedEntitiesKey = options.collection
        ? `selected${capitalize(options.collection)}Entities`
        : 'selectedEntities';
    const restoreFilterKey = options.collection
        ? `restore${capitalize(options.collection)}Filter`
        : 'restoreFilter';
    const updateFilterKey = options.collection
        ? `update${capitalize(options.collection)}Filter`
        : 'updateFilter';
    const updateSelectedKey = options.collection
        ? `updateSelected${capitalize(options.collection)}Entities`
        : 'updateSelected';
    const findKey = options.collection
        ? `find${capitalize(options.collection)}Entities`
        : 'find';
    const currentKey = options.collection
        ? `current${capitalize(options.collection)}`
        : 'current';
    const setCurrentKey = options.collection
        ? `setCurrent${capitalize(options.collection)}`
        : 'setCurrent';
    const insertKey = options.collection
        ? `insert${capitalize(options.collection)}`
        : 'insert';
    const updateKey = options.collection
        ? `update${capitalize(options.collection)}`
        : 'update';
    const updateAllByKey = options.collection
        ? `updateAll${capitalize(options.collection)}By`
        : 'updateAllBy';
    const removeKey = options.collection
        ? `remove${capitalize(options.collection)}`
        : 'remove';
    const removeAllByKey = options.collection
        ? `removeAll${capitalize(options.collection)}By`
        : 'removeAllBy';
    const entitiesKey = options.collection ? `${options.collection}Entities` : 'entities';
    const entityMapKey = options.collection ? `${options.collection}EntityMap` : 'entityMap';
    const idsKey = options.collection ? `${options.collection}Ids` : 'ids';
    const setQueryParamsKey = 'setQueryParams';
    const updateQueryParamsKey = 'updateQueryParams';
    return {
        filterKey,
        setQueryParamsKey,
        updateQueryParamsKey,
        selectedIdsKey,
        selectedEntitiesKey,
        restoreFilterKey,
        updateFilterKey,
        updateSelectedKey,
        findKey,
        entitiesKey,
        entityMapKey,
        idsKey,
        currentKey,
        setCurrentKey,
        insertKey,
        updateKey,
        updateAllByKey,
        removeKey,
        removeAllByKey,
    };
}
function withCollectionService(options) {
    let colService;
    const { collection: prefix, filter, query, countQuery } = options;
    const { filterKey, entitiesKey, selectedEntitiesKey, selectedIdsKey, setQueryParamsKey, updateQueryParamsKey, updateFilterKey, updateSelectedKey, currentKey, insertKey, updateKey, updateAllByKey, removeKey, removeAllByKey, setCurrentKey, } = getCollectionServiceKeys(options);
    const { callStateKey } = getCallStateKeys({ collection: prefix });
    const ensureWithEntities = (store) => {
        if (!(entitiesKey in store)) {
            throw new Error(`'withCollectionService' can only be used together with 'withEntities' from "@ngrx/singals/entities" signal store feature.`);
        }
    };
    const ensureCollection = () => {
        if (colService instanceof RxDBCollectionService)
            return;
        const injector = inject(Injector);
        colService = injector.get(RXDB_COLLECTION);
    };
    return signalStoreFeature(withState(() => {
        return {
            [filterKey]: filter,
            [selectedIdsKey]: {},
            [currentKey]: undefined,
            [queryKey]: query || {},
        };
    }), withComputed((store) => {
        ensureWithEntities(store);
        ensureCollection();
        const entities = store[entitiesKey];
        const selectedIds = store[selectedIdsKey];
        return {
            [selectedEntitiesKey]: computed(() => entities().filter(e => selectedIds()[e.id])),
            [countKey]: computed(() => entities().length),
            [countAllKey]: computedAsync(() => colService.count()),
            [`countFiltered`]: computedAsync(() => colService.count(countQuery)),
        };
    }), withMethods((store) => {
        ensureWithEntities(store);
        ensureCollection();
        return {
            ['sync']: async () => {
                store[callStateKey] && patchState(store, setLoading(prefix));
                await colService.sync();
                store[callStateKey] && patchState(store, setLoaded(prefix));
            },
            [setQueryParamsKey]: (q) => {
                colService.setQueryParams(q);
            },
            [updateQueryParamsKey]: (q) => {
                colService.patchQueryParams(q);
            },
            [updateFilterKey]: async (filter) => {
                store[callStateKey] && patchState(store, setLoading(prefix));
                if (typeof filter === 'string') {
                    await colService.setLocal(DEFAULT_LOCAL_DOCUMENT_ID, 'filter', filter);
                }
                else {
                    await colService.upsertLocal(DEFAULT_LOCAL_DOCUMENT_ID, filter);
                }
                patchState(store, { [filterKey]: filter });
                store[callStateKey] && patchState(store, setLoaded(prefix));
            },
            [updateSelectedKey]: (id, selected) => {
                patchState(store, (state) => ({
                    [selectedIdsKey]: {
                        ...state[selectedIdsKey],
                        [id]: selected,
                    },
                }));
            },
            [setCurrentKey]: (current) => {
                patchState(store, { [currentKey]: current });
            },
            [insertKey]: async (entity) => {
                store[callStateKey] && patchState(store, setLoading(prefix));
                try {
                    await colService.insert(entity);
                    store[callStateKey] && patchState(store, setLoaded(prefix));
                }
                catch (e) {
                    store[callStateKey] && patchState(store, setError(e, prefix));
                    throw e;
                }
            },
            [updateKey]: async (entity) => {
                store[callStateKey] && patchState(store, setLoading(prefix));
                try {
                    await colService.upsert(entity);
                    store[callStateKey] && patchState(store, setLoaded(prefix));
                }
                catch (e) {
                    store[callStateKey] && patchState(store, setError(e, prefix));
                    throw e;
                }
            },
            [updateAllByKey]: async (query, data) => {
                store[callStateKey] && patchState(store, setLoading(prefix));
                try {
                    await colService.updateBulk(query, data);
                    store[callStateKey] && patchState(store, setLoaded(prefix));
                }
                catch (e) {
                    store[callStateKey] && patchState(store, setError(e, prefix));
                    throw e;
                }
            },
            [removeKey]: async (entity) => {
                store[callStateKey] && patchState(store, setLoading(prefix));
                try {
                    await colService.remove(entity);
                    store[callStateKey] && patchState(store, setLoaded(prefix));
                }
                catch (e) {
                    store[callStateKey] && patchState(store, setError(e, prefix));
                    throw e;
                }
            },
            [removeAllByKey]: async (query) => {
                store[callStateKey] && patchState(store, setLoading(prefix));
                try {
                    await colService.removeBulk(query);
                    store[callStateKey] && patchState(store, setLoaded(prefix));
                }
                catch (e) {
                    store[callStateKey] && patchState(store, setError(e, prefix));
                    throw e;
                }
            },
        };
    }), withHooks({
        onInit: async (store) => {
            store[callStateKey] && patchState(store, setLoading(prefix));
            colService.queryParams$
                .pipe(tapOnce(() => {
                if (!isEmpty(query)) {
                    colService.setQueryParams(query);
                }
            }), tap(queryParams => {
                patchState(store, { query: queryParams });
                store[callStateKey] && patchState(store, setLoaded(prefix));
            }), switchMap(queryParams => colService.docs(queryParams)), tapResponse({
                next: result => {
                    store[callStateKey] && patchState(store, setLoaded(prefix));
                    return patchState(store, prefix
                        ? setAllEntities(result, { collection: prefix })
                        : setAllEntities(result));
                },
                error: e => {
                    logger.log(e);
                    store[callStateKey] && patchState(store, setError(e, prefix));
                },
                finalize: () => store[callStateKey] && patchState(store, setLoaded(prefix)),
            }))
                .subscribe();
        },
        onDestroy: () => {
            colService.destroy();
        },
    }));
}

export { getCollectionServiceKeys, withCollectionService };
//# sourceMappingURL=ngx-odm-rxdb-signals.mjs.map
