import { loadRxDBPlugins, RxDBService } from '@ngx-odm/rxdb/core';
import { randomCouchString, createRxDatabase } from 'rxdb';
import { getRxStorageMemory } from 'rxdb/plugins/storage-memory';

const TEST_SCHEMA = {
    type: 'object',
    title: 'Todo',
    description: 'Todo Schema',
    required: ['id', 'title', 'createdAt'],
    version: 0,
    encrypted: undefined,
    properties: {
        id: {
            type: 'string',
            pattern: '^(.*)$',
            maxLength: 32,
        },
        title: {
            type: 'string',
        },
        completed: {
            type: 'boolean',
        },
        createdAt: {
            type: 'number',
            multipleOf: 1,
            minimum: 0,
            maximum: Infinity,
        },
    },
    indexes: ['createdAt'],
    primaryKey: 'id',
};
const MOCK_DATA = [
    {
        id: '1',
        title: 'Test 1',
        completed: false,
        createdAt: 1546300800000,
    },
    {
        id: '2',
        title: 'Test 2',
        completed: true,
        createdAt: 1546300800100,
    },
    {
        id: '3',
        title: 'Test 3',
        completed: true,
        createdAt: 1546300800100,
    },
];
const MOCK_DATA_MAP = MOCK_DATA.reduce((acc, cur) => {
    acc[cur.id] = cur;
    return acc;
}, {});
const TEST_FEATURE_CONFIG_1 = {
    name: 'test',
    schema: TEST_SCHEMA,
    localDocuments: true,
    autoMigrate: true,
    options: {
        replicationStateFactory: undefined,
    },
};
const TEST_DB_CONFIG_1 = {
    name: 'test',
    storage: getRxStorageMemory(),
    multiInstance: false,
    ignoreDuplicate: true,
    localDocuments: true,
};
const TEST_DB_CONFIG_2 = {
    name: 'test',
    storage: getRxStorageMemory(),
    multiInstance: false,
    ignoreDuplicate: true,
    options: {
        storageType: 'dexie',
        schemas: {
            todo: { ...TEST_FEATURE_CONFIG_1 },
        },
    },
};
const getMockRxCollection = async (colConfig = TEST_FEATURE_CONFIG_1, randomName = false) => {
    await loadRxDBPlugins();
    const database = await createRxDatabase({
        name: randomName ? randomCouchString(6) : 'test',
        storage: getRxStorageMemory(),
        multiInstance: false,
        ignoreDuplicate: true,
        localDocuments: true,
    });
    const { test: collection } = await database.addCollections({
        [colConfig.name]: colConfig,
    });
    Object.getOwnPropertyNames(collection.__proto__).forEach(key => {
        if (typeof collection[key] === 'function') {
            jest.spyOn(collection, key);
        }
    });
    return collection;
};
const getMockRxdbService = async (colConfig = TEST_FEATURE_CONFIG_1, randomName = false) => {
    const collection = await getMockRxCollection(colConfig, randomName);
    const service = {
        db: null,
        collections: {},
        initDb: jest.fn(),
        destroyDb: jest.fn(),
        initCollections: jest.fn(),
    };
    jest.spyOn(service, 'initDb').mockImplementation(() => {
        service.db = collection.database;
        return Promise.resolve();
    });
    jest.spyOn(service, 'destroyDb').mockImplementation(() => {
        service.db = null;
    });
    jest.spyOn(service, 'initCollections').mockImplementation(() => {
        service.db = collection.database;
        service.collections[colConfig.name] = collection;
        return Promise.resolve(service.collections);
    });
    Object.setPrototypeOf(service, RxDBService.prototype);
    return service;
};

function setupNavigationWarnStub() {
    const warn = console.warn;
    const error = console.error;
    jest.spyOn(console, 'warn').mockImplementation((...args) => {
        const [firstArg] = args;
        if (typeof firstArg === 'string' &&
            firstArg.startsWith('Navigation triggered outside Angular zone')) {
            return;
        }
        return warn.apply(console, args);
    });
    jest.spyOn(console, 'error').mockImplementation((...args) => {
        const [firstArg] = args;
        if (typeof firstArg === 'string' && firstArg.startsWith('Attempted to log "[DEBUG')) {
            return;
        }
        return error.apply(console, args);
    });
}

export { MOCK_DATA, MOCK_DATA_MAP, TEST_DB_CONFIG_1, TEST_DB_CONFIG_2, TEST_FEATURE_CONFIG_1, TEST_SCHEMA, getMockRxCollection, getMockRxdbService, setupNavigationWarnStub };
//# sourceMappingURL=ngx-odm-rxdb-testing.mjs.map
