import { prepareQuery } from 'rxdb';
import { RxReplicationState } from 'rxdb/plugins/replication';
import { tap, timer, retry, defer, switchMap, map } from 'rxjs';

var NgxRxdbUtils;
(function (NgxRxdbUtils) {
    function clone(value) {
        if (Array.isArray(value)) {
            return value.slice();
        }
        else if (NgxRxdbUtils.isObject(value)) {
            return { ...value };
        }
        else {
            return value;
        }
    }
    NgxRxdbUtils.clone = clone;
    function keys(object) {
        let val = keysOfNonArray(object);
        if (Array.isArray(object)) {
            val = val.filter(item => item !== 'length');
        }
        return val;
    }
    NgxRxdbUtils.keys = keys;
    function keysOfNonArray(object) {
        return object ? Object.getOwnPropertyNames(object) : [];
    }
    NgxRxdbUtils.keysOfNonArray = keysOfNonArray;
    function forOwn(object, iteratee) {
        forEachOfArray(keys(object), key => iteratee(object[key], key));
        return object;
    }
    NgxRxdbUtils.forOwn = forOwn;
    function forOwnOfNonArray(object, iteratee) {
        forEachOfArray(keysOfNonArray(object), key => iteratee(object[key], key));
        return object;
    }
    NgxRxdbUtils.forOwnOfNonArray = forOwnOfNonArray;
    function forEach(collection, iteratee) {
        if (Array.isArray(collection)) {
            forEachOfArray(collection, iteratee);
        }
        else {
            forOwnOfNonArray(collection, iteratee);
        }
        return collection;
    }
    NgxRxdbUtils.forEach = forEach;
    function forEachOfArray(array, iteratee) {
        for (let i = 0, len = array.length; i < len; ++i) {
            if (iteratee(array[i], i) === false) {
                break;
            }
        }
    }
    NgxRxdbUtils.forEachOfArray = forEachOfArray;
    function merge(object, ...sources) {
        for (const source of sources) {
            forEach(source, (value, key) => {
                const myValue = object[key];
                if (myValue instanceof Object) {
                    value = merge(clone(myValue), value);
                }
                object[key] = value;
            });
        }
        return object;
    }
    NgxRxdbUtils.merge = merge;
    function isEmpty(value) {
        if (!Array.isArray(value)) {
            value = keysOfNonArray(value);
        }
        return value.length === 0;
    }
    NgxRxdbUtils.isEmpty = isEmpty;
    NgxRxdbUtils.isFunction = (value) => typeof value === 'function';
    NgxRxdbUtils.isUndefined = (value) => value === undefined || value === 'undefined';
    function isNullOrUndefined(value) {
        return value === undefined || value === null || value === 'undefined' || value === 'null';
    }
    NgxRxdbUtils.isNullOrUndefined = isNullOrUndefined;
    NgxRxdbUtils.isObject = (x) => Object.prototype.toString.call(x) === '[object Object]';
    NgxRxdbUtils.isEmptyObject = (x) => isNullOrUndefined(x) || (NgxRxdbUtils.isObject(x) && isEmpty(x));
    function isDevMode() {
        return typeof globalThis['ngDevMode'] === 'undefined' || !!globalThis['ngDevMode'];
    }
    NgxRxdbUtils.isDevMode = isDevMode;
    function isValidNumber(value) {
        return typeof value === 'number' && !isNaN(value);
    }
    NgxRxdbUtils.isValidNumber = isValidNumber;
    function noop() {
        return void 0;
    }
    NgxRxdbUtils.noop = noop;
    function identity(value) {
        return value;
    }
    NgxRxdbUtils.identity = identity;
    function times(n, iteratee) {
        const result = [];
        for (let i = 0; i < n; ++i) {
            result[i] = iteratee(i);
        }
        return result;
    }
    NgxRxdbUtils.times = times;
    function range(start, end, step) {
        if (NgxRxdbUtils.isUndefined(end)) {
            end = start;
            start = 0;
        }
        if (NgxRxdbUtils.isUndefined(step)) {
            step = end < start ? -1 : 1;
        }
        return times(Math.abs((end - start) / (step || 1)), i => start + step * i);
    }
    NgxRxdbUtils.range = range;
    function getMaybeId(entityOrId) {
        if (NgxRxdbUtils.isObject(entityOrId)) {
            return entityOrId['_id'];
        }
        return String(entityOrId);
    }
    NgxRxdbUtils.getMaybeId = getMaybeId;
    function compact(array) {
        return array.filter(identity);
    }
    NgxRxdbUtils.compact = compact;
    function compactObject(obj) {
        if (isEmpty(obj)) {
            return obj;
        }
        if (Array.isArray(obj)) {
            return compact(obj);
        }
        return Object.entries(obj)
            .filter(([, value]) => !isNullOrUndefined(identity(value)) && !isEmpty(value))
            .reduce((acc, [key, val]) => ({ ...acc, [key]: val }), {});
    }
    NgxRxdbUtils.compactObject = compactObject;
    function isDevModeForced() {
        return localStorage['debug']?.includes(`@ngx-odm/rxdb`);
    }
    NgxRxdbUtils.isDevModeForced = isDevModeForced;
    function isTestEnvironment() {
        return ((typeof __karma__ !== 'undefined' && !!__karma__) ||
            (typeof jasmine !== 'undefined' && !!jasmine) ||
            (typeof jest !== 'undefined' && !!jest) ||
            (typeof Mocha !== 'undefined' && !!Mocha));
    }
    NgxRxdbUtils.isTestEnvironment = isTestEnvironment;
    NgxRxdbUtils.logger = {
        log: (function () {
            const bgColor = '#8d2089';
            if (isTestEnvironment() || !isDevMode() || !isDevModeForced()) {
                return noop;
            }
            return console.log.bind(console, `%c[DEBUG::@ngx-odm/rxdb]`, `background:${bgColor};color:#fff;padding:2px;font-size:normal;`);
        })(),
        table: (function () {
            if (isTestEnvironment() || !isDevMode() || !isDevModeForced()) {
                return noop;
            }
            return console.table.bind(console);
        })(),
    };
    function debug(tag = 'Event') {
        return function debugFn(source) {
            if (!isDevMode()) {
                return source;
            }
            return source.pipe(tap({
                next(value) {
                    NgxRxdbUtils.logger.log(`observable:${tag}:next:`, value);
                },
                error(e) {
                    NgxRxdbUtils.logger.log(`observable:${tag}:error:`, e.message);
                },
                complete() {
                    NgxRxdbUtils.logger.log(`observable:${tag}:complete`);
                },
            }));
        };
    }
    NgxRxdbUtils.debug = debug;
    function retryWithBackoff(count = 3, retryTime = 10000) {
        return (obs$) => obs$.pipe(retry({
            count,
            delay: (_, retryIndex) => {
                const d = Math.pow(2, retryIndex - 1) * retryTime;
                NgxRxdbUtils.logger.log('replication:kinto:longpoll:retry', retryIndex, d);
                return timer(d);
            },
        }));
    }
    NgxRxdbUtils.retryWithBackoff = retryWithBackoff;
    NgxRxdbUtils.tapOnce = (callback) => {
        return (source) => defer(() => {
            callback();
            return source;
        });
    };
    NgxRxdbUtils.deferUntil = (subj) => {
        return (source) => defer(() => {
            return subj.pipe(switchMap(() => source));
        });
    };
    NgxRxdbUtils.getDefaultQuery = () => ({
        selector: { _deleted: { $eq: false } },
        skip: 0,
        sort: [{ id: 'asc' }],
    });
    const _schema = {
        properties: {
            id: { type: 'string' },
            _deleted: { type: 'boolean' },
        },
        primaryKey: 'id',
        indexes: [['_deleted', 'id']],
    };
    const _queryPlan = {
        query: NgxRxdbUtils.getDefaultQuery(),
        queryPlan: {
            index: ['_deleted', 'id'],
            startKeys: [false, -9007199254740991],
            endKeys: [false, '￿'],
            inclusiveEnd: true,
            inclusiveStart: true,
            sortSatisfiedByIndex: true,
            selectorSatisfiedByIndex: false,
        },
    };
    NgxRxdbUtils.getDefaultPreparedQuery = () => prepareQuery(_schema, NgxRxdbUtils.getDefaultQuery());
})(NgxRxdbUtils || (NgxRxdbUtils = {}));
const getDefaultFetch = () => {
    if (typeof window === 'object' && 'fetch' in window) {
        return window.fetch.bind(window);
    }
    else {
        return fetch;
    }
};
function getDefaultFetchWithHeaders(headers = {}) {
    const fetch = getDefaultFetch();
    const ret = (url, options = {}) => {
        Object.assign(options, {
            headers: {
                ...(options.headers || {}),
                ...headers,
            },
        });
        return fetch(url, options);
    };
    return ret;
}
function isValidRxReplicationState(obj) {
    return obj && obj instanceof RxReplicationState;
}
function mapFindResultToJsonArray(withRevAndAttachments = false) {
    return map(docs => {
        return (Array.isArray(docs) ? docs : [...docs.values()]).map(d => {
            const data = { ...d._data };
            if (!withRevAndAttachments) {
                delete data._rev;
                delete data._attachments;
                delete data._deleted;
                delete data._meta;
            }
            return data;
        });
    });
}

export { NgxRxdbUtils, getDefaultFetch, getDefaultFetchWithHeaders, isValidRxReplicationState, mapFindResultToJsonArray };
//# sourceMappingURL=ngx-odm-rxdb-utils.mjs.map
